from ..base_rest import BaseRest

class Intraday(BaseRest):
    def products(self, **params):
        return self.request(f"intraday/products", **params)
    
    def tickers(self, **params):
        return self.request(f"intraday/tickers", **params)
    
    def ticker(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"intraday/ticker/{symbol}", **params)
    
    def quote(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"intraday/quote/{symbol}", **params)
    
    def candles(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"intraday/candles/{symbol}", **params)
    
    def trades(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"intraday/trades/{symbol}", **params)
    
    def volumes(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"intraday/volumes/{symbol}", **params)


from .intraday import Intraday
from .historical import Historical



class RestFutOptClient:
    def __init__(self, **config):
        # config: base_url, api_key?, bearer_token?
        self.config = config

    @property
    def intraday(self):
        return Intraday(**self.config)
    
    @property
    def historical(self):
        return Historical(**self.config)


from urllib.parse import urlencode
import requests
from ..exceptions import FugleAPIError


class BaseRest(object):
    def __init__(self, **config):
        self.config = config

    def request(self, path, **params):
        baseUrl = self.config['base_url']
        headers = {}
        if self.config.get('api_key'):
            headers['X-API-KEY'] = self.config['api_key']
        if self.config.get('bearer_token'):
            headers['Authorization'] = f"Bearer {self.config['bearer_token']}"
        if self.config.get('sdk_token'):
            headers['X-SDK-TOKEN'] = self.config['sdk_token']

        endpoint = path if (path.startswith('/')) else '/' + path

        if len(params) == 0:
            query = ''
        else:
            query = '?' + urlencode(params)

        url = baseUrl + endpoint + query

        try:
            response = requests.get(url, headers=headers)

            # 檢查 HTTP 錯誤狀態
            if response.status_code >= 400:
                error_msg = f"HTTP {response.status_code}"
                try:
                    error_data = response.json()
                    if 'message' in error_data:
                        error_msg = error_data['message']
                except:
                    pass

                raise FugleAPIError(
                    error_msg,
                    url=url,
                    status_code=response.status_code,
                    params=params,
                    response_text=response.text
                )

            return response.json()

        except ValueError as e:
            raise FugleAPIError(
                "Failed to parse JSON response",
                url=url,
                status_code=response.status_code,
                params=params,
                response_text=response.text
            )

        except requests.exceptions.RequestException as e:
            raise FugleAPIError(
                f"{type(e).__name__}: {str(e)}",
                url=url,
                params=params
            )



from ..base_rest import BaseRest


class CorporateActions(BaseRest):
    def capital_changes(self, **params):
        return self.request("corporate-actions/capital-changes", **params)

    def dividends(self, **params):
        return self.request("corporate-actions/dividends", **params)

    def listing_applicants(self, **params):
        return self.request("corporate-actions/listing-applicants", **params)



from ..client_factory import ClientFactory
from .futopt import WebSocketFutOptClient
from .stock import WebSocketStockClient
from ..constants import FUGLE_MARKETDATA_API_WEBSOCKET_BASE_URL, FUGLE_MARKETDATA_API_VERSION

class WebSocketClientFactory(ClientFactory):
    def __init__(self, **options):
        super().__init__(**options)
        self.__clients = {}
        self.options = options

    @property
    def stock(self):
        return self.get_client('stock')

    @property
    def futopt(self):
        return self.get_client('futopt')

    def get_client(self, type):
        base_url = self.options.get('base_url')
        if not base_url:
            base_url = f"{FUGLE_MARKETDATA_API_WEBSOCKET_BASE_URL}/{FUGLE_MARKETDATA_API_VERSION}"
        
        url = f'{base_url.rstrip("/")}/{type}/streaming'

        if type in self.__clients:
            return self.__clients[type]
        
        client_options = {**self.options, 'base_url': url}

        if type == 'stock': 
            client = WebSocketStockClient(**client_options)
        elif type == 'futopt' :
            client = WebSocketFutOptClient(**client_options)
        else: 
            None

        self.__clients[type] = client
        return client








from ..base_rest import BaseRest

class Snapshot(BaseRest):
    def quotes(self, **params):
        market = params.pop('market')
        return self.request(f"snapshot/quotes/{market}", **params)
    
    def movers(self, **params):
        market = params.pop('market')
        return self.request(f"snapshot/movers/{market}", **params)
    
    def actives(self, **params):
        market = params.pop('market')
        return self.request(f"snapshot/actives/{market}", **params)
    


from ..base_rest import BaseRest

class Historical(BaseRest):
    def candles(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"historical/candles/{symbol}", **params)
    
    def stats(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"historical/stats/{symbol}", **params)
    


import orjson
import websocket
from typing import Optional
from pyee import EventEmitter
from threading import Thread, Timer

from ..constants import (
    AUTHENTICATION_TIMEOUT_MESSAGE,
    CONNECT_EVENT,
    DISCONNECT_EVENT,
    MESSAGE_EVENT,
    ERROR_EVENT,
    AUTHENTICATED_EVENT,
    MISSING_CREDENTIALS_MESSAGE,
    UNAUTHENTICATED_EVENT,
    UNAUTHENTICATED_MESSAGE
)

websocket.setdefaulttimeout(5)


class HealthCheckConfig:
    def __init__(self, enabled: bool = False, ping_interval: int = 30000, max_missed_pongs: int = 2):
        self.enabled = enabled
        self.ping_interval = ping_interval
        self.max_missed_pongs = max_missed_pongs


class AuthenticationState:
    PENDING = 0
    AUTHENTICATING = 1
    AUTHENTICATED = 2
    UNAUTHENTICATED = 3


class WebSocketClient():
    def __init__(self, **config):
        self.config = config
        self.health_check: Optional[HealthCheckConfig] = config.get('health_check')
        self.ee = EventEmitter()
        self.ee.on(CONNECT_EVENT, self.__authenticate)
        self.__ws = websocket.WebSocketApp(
            self.config.get('base_url'),
            on_open=self.__on_open,
            on_close=self.__on_close,
            on_error=self.__on_error,
            on_message=self.__on_message)

        self.auth_timer = None
        self.auth_status = AuthenticationState.PENDING
        self.error = None

        # Health check properties
        self.ping_timer = None
        self.missed_pongs = 0

    def ping(self, message):
        message = {
            "event": "ping",
            "data": {
                "state": message
            }
        }
        self.__send(message)

    def subscribe(self, params):
        message = {
            "event": "subscribe",
            "data": params
        }
        self.__send(message)

    def unsubscribe(self, params):
        message = {
            "event": "unsubscribe",
            "data": params
        }
        self.__send(message)

    def subscriptions(self):
        message = {
            "event": "subscriptions"
        }
        self.__send(message)

    def __authenticate(self):
        if self.config.get('api_key'):
            auth_info = {
                'event': 'auth',
                'data': {
                    'apikey': self.config['api_key']
                }
            }
        elif self.config.get('bearer_token'):
            auth_info = {
                'event': 'auth',
                'data': {
                    'token': self.config['bearer_token']
                }
            }
        elif self.config.get('sdk_token'):
            auth_info = {
                'event': 'auth',
                'data': {
                    'sdkToken': self.config['sdk_token']
                }
            }
        else:
            self.auth_status = AuthenticationState.UNAUTHENTICATED
            self.error = Exception(MISSING_CREDENTIALS_MESSAGE)

        self.__send(auth_info)
        self.auth_status = AuthenticationState.AUTHENTICATING
        self.auth_timer = Timer(5, self.check_auth_status)
        self.auth_timer.start()

    def __send(self, message):
        self.__ws.send(orjson.dumps(message).decode('utf-8'))

    def __on_open(self, ws):
        self.ee.emit(CONNECT_EVENT)

    def __on_close(self, ws, close_status_code, close_msg):
        self.ee.emit(DISCONNECT_EVENT, close_status_code, close_msg)

    def __on_message(self, ws, data):
        message = orjson.loads(data)
        self.ee.emit(MESSAGE_EVENT, data)
        if message['event'] == AUTHENTICATED_EVENT:
            self.ee.emit(AUTHENTICATED_EVENT, message)
            self.auth_status = AuthenticationState.AUTHENTICATED

            # Start health check if enabled
            if self.health_check and self.health_check.enabled:
                self.__start_health_check()
        elif message['event'] == ERROR_EVENT:
            if message['data'] and message['data']['message'] == UNAUTHENTICATED_MESSAGE:
                self.ee.emit(UNAUTHENTICATED_EVENT, message)
                self.auth_status = AuthenticationState.UNAUTHENTICATED
                self.error = Exception(UNAUTHENTICATED_MESSAGE)
        elif message['event'] == 'pong':
            # Reset missed pongs counter
            self.missed_pongs = 0

    def __on_error(self, ws, error):
        self.ee.emit(ERROR_EVENT, error)

    def on(self, event, listener):
        self.ee.on(event, listener)

    def off(self, event, listener):
        self.ee.off(event, listener)

    def check_auth_status(self):
        if self.auth_status == AuthenticationState.AUTHENTICATING:
            self.auth_status = AuthenticationState.UNAUTHENTICATED
            self.error = Exception(AUTHENTICATION_TIMEOUT_MESSAGE)

    def __start_health_check(self):
        """Start the health check ping/pong mechanism"""
        if self.health_check and self.health_check.enabled:
            self.missed_pongs = 0
            self.__send_ping()

    def __send_ping(self):
        """Send ping and schedule next ping"""
        if not self.health_check or not self.health_check.enabled:
            return

        try:
            self.ping("")
            self.missed_pongs += 1
            self.__check_missed_pongs()

            # Schedule next ping
            interval_seconds = self.health_check.ping_interval / 1000.0
            self.ping_timer = Timer(interval_seconds, self.__send_ping)
            self.ping_timer.start()
        except Exception as error:
            print(f"Failed to send ping: {error}")
            self.disconnect()

    def __check_missed_pongs(self):
        """Check if too many pongs have been missed"""
        if self.health_check and self.health_check.enabled:
            if self.missed_pongs > self.health_check.max_missed_pongs:
                self.disconnect()
                raise Exception(f"Did not receive pong for {self.health_check.max_missed_pongs} consecutive times. Disconnecting...")

    def connect(self):
        Thread(target=self.__ws.run_forever).start()
        while True:
            if self.auth_status in [AuthenticationState.AUTHENTICATED, AuthenticationState.UNAUTHENTICATED]:
                break
        if self.error is not None:
            self.__ws.close()
            self.auth_timer.cancel()
            raise self.error

    def disconnect(self):
        if self.__ws is not None:
            self.__ws.close()

        if self.auth_timer is not None:
            self.auth_timer.cancel()
            self.auth_timer = None

        if self.ping_timer is not None:
            self.ping_timer.cancel()
            self.ping_timer = None

        self.auth_status = AuthenticationState.PENDING
        self.error = None



from ..client_factory import ClientFactory
from ..constants import FUGLE_MARKETDATA_API_REST_BASE_URL, FUGLE_MARKETDATA_API_VERSION
from .stock import RestStockClient
from .futopt import RestFutOptClient


class RestClientFactory(ClientFactory):
    def __init__(self, **options):
        super().__init__(**options)
        self.__clients = {}
        self.options = options

    @property
    def stock(self):
        return self.get_client('stock')

    @property
    def futopt(self):
        return self.get_client('futopt')

    def get_client(self, type):
        base_url = self.options.get('base_url')
        if not base_url:
            base_url = f"{FUGLE_MARKETDATA_API_REST_BASE_URL}/{FUGLE_MARKETDATA_API_VERSION}"

        url = f'{base_url.rstrip("/")}/{type}'

        if type in self.__clients:
            return self.__clients[type]

        # Create a copy of options and override base_url
        client_options = {**self.options}
        client_options['base_url'] = url

        if type == 'stock':
            client = RestStockClient(**client_options)

        elif type == 'futopt':
            client = RestFutOptClient(**client_options)

        else:
            None

        self.__clients[type] = client
        return client



class ClientFactory:
    def __init__(self, **options):
        self.options = options
        api_key = options.get('api_key')
        bearer_token = options.get('bearer_token')
        sdk_token = options.get('sdk_token')
        token_count = sum(bool(token) for token in [api_key, bearer_token, sdk_token])

        if token_count == 0:
            raise TypeError('One of the "apiKey", "bearerToken", or "sdkToken" options must be specified')

        if token_count > 1:
            raise TypeError('Only one of the "apiKey", "bearerToken", or "sdkToken" options must be specified')



class FugleAPIError(Exception):
    """Fugle API 錯誤，包含 debug 資訊"""

    def __init__(self, message, url=None, status_code=None, params=None, response_text=None):
        self.message = message
        self.url = url
        self.status_code = status_code
        self.params = params
        self.response_text = response_text

        # 建構完整的錯誤訊息
        error_parts = [f"[Fugle API Error] {message}"]
        if url:
            error_parts.append(f"URL: {url}")
        if status_code:
            error_parts.append(f"Status: {status_code}")
        if params:
            error_parts.append(f"Params: {params}")
        if response_text:
            error_parts.append(f"Response: {response_text[:200]}...")

        super().__init__('\n'.join(error_parts))



from .rest import RestClientFactory as RestClient
from .websocket import WebSocketClientFactory as WebSocketClient, HealthCheckConfig
from .exceptions import FugleAPIError

__version__ = '2.4.1'

__all__ = ['RestClient', 'WebSocketClient', 'HealthCheckConfig', 'FugleAPIError', '__version__']



from .intraday import Intraday
from .historical import Historical
from .snapshot import Snapshot
from .technical import Technical
from .corporate_actions import CorporateActions


class RestStockClient:
    def __init__(self, **config):
        # config: base_url, api_key?, bearer_token?
        self.config = config

    @property
    def intraday(self):
        return Intraday(**self.config)
    
    @property
    def historical(self):
        return Historical(**self.config)
    
    @property
    def snapshot(self):
        return Snapshot(**self.config)
    
    @property
    def technical(self):
        return Technical(**self.config)

    @property
    def corporate_actions(self):
        return CorporateActions(**self.config)



FUGLE_MARKETDATA_API_REST_BASE_URL = 'https://api.fugle.tw/marketdata'
FUGLE_MARKETDATA_API_WEBSOCKET_BASE_URL = 'wss://api.fugle.tw/marketdata'
FUGLE_MARKETDATA_API_VERSION = 'v1.0'

CONNECT_EVENT = 'connect'
DISCONNECT_EVENT = 'disconnect'
MESSAGE_EVENT = 'message'
ERROR_EVENT = 'error'
AUTHENTICATED_EVENT = 'authenticated'
UNAUTHENTICATED_EVENT = 'unauthenticated'
UNAUTHENTICATED_MESSAGE = 'Invalid authentication credentials'
AUTHENTICATION_TIMEOUT_MESSAGE = 'authentication timeout'
MISSING_CREDENTIALS_MESSAGE= 'missing authentication credentials'



from ..base_rest import BaseRest

class Technical(BaseRest):
    def sma(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"technical/sma/{symbol}", **params)

    def rsi(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"technical/rsi/{symbol}", **params)

    def kdj(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"technical/kdj/{symbol}", **params)

    def macd(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"technical/macd/{symbol}", **params)

    def bb(self, **params):
        symbol = params.pop('symbol')
        return self.request(f"technical/bb/{symbol}", **params)
